% This file demonstrates how the models developed by Tamal Das can be used
% for steady state optimization; in this case we demonstrate steady state finsing problem. 
% The system includes 1st stage separator, 2nd stage separator, 
% hydrocyclone and compact flotation unit. All these models are connected one after the other.  

% Download the casadi library folder for MATLAB from CasADi website

clear; clc;
addpath('C:\Users\NOTADAS\Downloads\casadi-windows-matlabR2016a-v3.4.5'); % Provide the path of the Casadi folder here
import casadi.*
init_param;  % parameter initialization
 
%% Definition of CasADi variables :
ns_gs1 = 543;   %  number of states in 1st stage separator model
ns_hc = 432;    %  number of states in hydrocyclone model
ns_cfu = 35;    %  number of states in compact flotation unit model
ns_tot = ns_gs1+ns_hc+ns_cfu;   
nu_gs1 = 1;   %  number of control inputs in 1st stage separator model, we here use only gas flow because others (water and oil flows) are calculated based on levels (water level oil level). 

% in gravity separators (1st stage and 2nd stage separators), two controlled variables are oil level and water level, which are integrating states
% The dynamic model hence cannot run in open loop, however with controllers
% for the two level one could run the dynamic model. So what do we do when
% we want to use the same model in steady state or steady state analysis,
% well we turn off the equations for derivatives of the two levels by
% multiplying the right side of the equation with 0. And we make the
% optimizer to use either the levels or the flows as decision variables. 
% Here we make the optimizer find the levels and not the flows. Since the oil
% and water flows depend on the oil and water levels, given those levels
% the flows can be internally calculated hence we cannot expect optimizer
% to find levels and flows independently. We tried with the flows as
% decision variables, but it was causing some issues. See the casADi_GS1 and casADi_GS2
% files to see where we multiply with 0.

nu_hc = 1;      %  number of control inputs in hydrocyclone model
nu_cfu = 3;     %  number of control inputs in compact flotation unit model, reject flow, flotation gas flow, water outflow
nu_tot = nu_gs1+nu_hc+nu_cfu;
nd = 3;         %  number of disturbances: gas flow, total liquid flow, water cut in the liquid flow

%% Solve Optimal Bulk separation:

X       = SX.sym('X',ns_gs1+ns_hc+ns_cfu,1);                     % Casadi definition of States    
X_dot   = SX.sym('X_dot',ns_gs1+ns_hc+ns_cfu,1);                 % Casadi definition of State derivatives    
U       = SX.sym('U',nu_gs1+nu_hc+nu_cfu,1);                   % Casadi definition of control inputs 
D       = SX.sym('D',nd,1);                       % Casadi definition of Disturbances

% Definition of CasADi Model equations and functions:
casADi_GS1;  % Import the casadi model of 1st stage separator
casADi_HC;   % Import the casadi model of hydrocyclone
casADi_CFU;  % Import the casadi model of compact flotation unit

dots = [dots_gs1;dots_hc;dots_cfu];   % Collecting all the state derivatives from different models in a vector
output = [output_gs1; output_hc; output_cfu];  % Collecting all the outputs from different models in a vector
f = Function('f', {X,U,D}, {dots});     % Creating a casadi function of all state derivatives from different models
y = Function('y', {X,U,D}, {output});   % Creating a casadi function of all outputs from different models 

problem.f = 0;   % Objective function of the optimization problem   
problem.x = [X;U];  % Decision variables for the optimization problem include states and control inputs 
problem.g = [f(X,U,D);output(10);output(end-2:end)]; % Other variables that we want to constaint in the optimization problem
problem.p = D;  % Parameters for the optimization problem are disturbances to the system
ipopt_opt.ipopt.acceptable_tol = 1e-8; % Tolerance of the optimization problem
solverss = nlpsol('solverss','ipopt',problem,ipopt_opt);  % Creating the steady state finding solver for steady state finding, it might take some time to create this solver so dont panic

%% ss
x_lb = [zeros(ns_gs1-3,1); 2.2; 1.3342; 8;    % gs1
        zeros(ns_hc,1);...%];                                     % hc
        zeros(ns_cfu-5,1); 0;  0.9 ; 10/1e16/(pi*(25e-6)^3/6) ; 0; 0] ;                  % cfu  % The lower bounds of states

x_ub = [inf*ones(ns_gs1-3,1);  2.2; 1.3342; 8; % gs1
        inf*ones(ns_hc,1); ... %];                                  % hc
        inf*ones(ns_cfu-5,1); 1; 1.1 ; 10.1/1e16/(pi*(25e-6)^3/6) ; 1; 1] ;         % cfu    % The upper bounds of states

     
%D_val(2) = D_val(2)*1.1;      
%D_val(3) = D_val(3)*1.1;     
    
g_lb = [zeros(ns_gs1,1);zeros(ns_hc,1);zeros(ns_cfu,1);0;0.54;0;0];        % The lower bounds of variables in problem.g
g_ub = [zeros(ns_gs1,1);zeros(ns_hc,1);zeros(ns_cfu,1);100;0.55;1;1];       % The upper bounds of variables in problem.g

% load('SS_solution')
% x0 = [SS_sol_X2_95_951(1:450);zeros(90,1);SS_sol_X2_95_951(451:453);SS_sol_X2_95_951(453+1:453+450);zeros(90,1);SS_sol_X2_95_951(453+450+1:end)]; 
% U0 = SS_sol_U2_95_951;  % loading an initial guess for the decision variables

load('tempSol'); U0 =U_sol;

u_lb = [0;2/100; 0; 0; 0]; %u_ub = [U_val(1);0.5/100;25;D_val(2)*D_val(3)*3600/4; 33]; % lower and upper bounds of the control inputs
u_ub = [inf;2/100;inf;inf;inf]; % lower and upper bounds of the control inputs

%%
x0(543+1:543+542) = [];
sol=solverss( 'p',D_val,'x0',[x0;U0],'lbx',[x_lb;u_lb],...
            'ubx',[x_ub;u_ub],'lbg',g_lb,'ubg',g_ub);    % calling the solver to solve the problem
      
x_sol = full(sol.x);  % extracting the solution
X_sol = x_sol(1:ns_gs1+ns_hc+ns_cfu); 
U_sol = x_sol(ns_gs1+ns_hc+ns_cfu+1:ns_gs1+ns_hc+ns_cfu+5)

x0 = X_sol; % U0  = U_sol

% save('tempSol1','x0','U_sol');

output_sol =  y(X_sol,U_sol,D_val)
output_sol =  full(y(X_sol,U_sol,D_val));
solution.gs1_hl  = x0(ns_gs1-2);
solution.gs1_hlw = x0(ns_gs1-1);
solution.gs1_P   = x0(ns_gs1);
solution.gs_ppm_out = output_sol(6);
solution.hc_ppm_out = output_sol(8);
solution.hc_eff = 1-output_sol(8)/output_sol(6);
solution.hc_flowsplit_percent = output_sol(9)/(output_sol(7)+output_sol(9))*100;
solution.cfu_ppm_out = output_sol(16);
solution.cfu_eff = 1-output_sol(16)/output_sol(8);
solution.cfu_P = output_sol(14);
solution.cfu_F_out = output_sol(15);
solution.cfu_alpha_l = x0(ns_gs1+ns_hc+ns_cfu-4)